/*
 * pppdwrap (C) 1999 Andrew J. Kroll (GPL) See file COPYING
 
 
			     ______________
			    /\            /\ 
			   /  \  modems  /  \
			  /    \        /    \
			 / pppd \__on__/ wrap \ 
			 \      /      \      /
			  \    /        \    /
			   \  / networks \  /
			    \/____________\/


Special thanks to Alan Cox for explaining tty/pty's and how they
interconnect, as well as a few of his code snippits...
(of which I shamelessly stole.)

*/


#define PORT 6969

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef HPUX
#include <strings.h>
#endif

#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <fcntl.h>
#include <sys/file.h>
#include <arpa/inet.h>

#define POSIX
#define TERMIO termios
#include <termios.h>

#define TELOPTS
#include <arpa/telnet.h>

#ifdef UNIX98_YUCK
#include <bits/stropts.h>

#endif


struct progmode {
	unsigned short int passall;
	unsigned short int forked;
	unsigned short int recog;
	unsigned short int passflag;
	unsigned short int progmode;
	unsigned short int control;
};

struct sockaddr_in server;
int sock=0,outflag=0,inflag=0;
pid_t child=0,leader=0;

int fdPTY,fdTTY;
struct progmode pm;
time_t start;
int flag=1;

#ifdef DEBUG
FILE *foo;
#endif

#include "common.h"

void instrip(int fd,unsigned char *buf,int length,unsigned int strip)
{
	int ret;

	do { 
		ret = write(fd, buf, length); 
	} while( ret == -1 && errno==EINTR);
}

void outpad(int fd,unsigned char *buf,int length,unsigned int pad)
{
	int ret;
	do { 
		ret = write(fd, buf, length); 
	} while( ret == -1 && errno==EINTR);
}

void problemo(char *string)
{
	pid_t me;
	int try=0;
	time_t logtime;

	me=getpid();
	if(sock) close(sock);
	if(child!=me && child) {
		kill(child,1);
		sleep(5);
		while(kill(child,0)!=-1 && try++<5) {
			kill(child,1);
			sleep(2);
		}
		try=0;
		while(kill(child,0)!=-1 && try++<5) {
			kill(child,15);
			sleep(2);
		}
		try=0;
		while(kill(child,0)!=-1 && try++<5) {
			kill(child,9);
			sleep(2);
		}

	}
	if(child!=me) {
		time(&logtime);
		printf("[%s] %s\n",strpn(ctime(&logtime)),string);
		fflush(stdout);
		fflush(stdout);
	}
	exit(0);
}


void fireman() {
  char crap[160];
  union wait status;
  while(wait3(&status,WNOHANG,(struct rusage *)NULL)>0);
  sprintf(crap, "Child process terminated - executed for %lu seconds\r\n",
	  (unsigned long int)time(NULL)-start);
  problemo(crap);
  pm.forked = 0;
}


void cclosed() {
  problemo("Connection closed by foreign host.");
#ifdef DEBUG
  fclose(foo);
#endif

}



#include "snag.h"

unsigned long int resolver(host)
char *host;
{
	int x=0;
	struct hostent *he;

	if(sscanf(host,"%d.%d.%d.%d",&x,&x,&x,&x)==4 || !strcmp(host,"0"))
		return(inet_addr(host));
	while(!(he=gethostbyname((char *)host)) && x++<3)
		sleep(1);
	if(x<3)
		return(htonl((unsigned long int)((unsigned char)he->h_addr_list[0][0]*
			(unsigned int)256+(unsigned char)he->h_addr_list[0][1])*
			(unsigned int)65536+(unsigned long int)((unsigned char)
			he->h_addr_list[0][2]*(unsigned int)256+(unsigned char)
			he->h_addr_list[0][3])));

	problemo("Unable to resolve host name.");
	return -1;
}

int mypopen(int *pty,int *tty,char *arguments)
{

/* this gives a pedantic error, anyone really care to clean this one up? */

	char *argp[] = { "/bin/bash", "-c", NULL, NULL };
	argp[2]=arguments;

#ifdef DEBUG
	fprintf(stderr,"/bin/bash -c %s\r\n", argp[2]);
#endif
 /* the tty side you stick pppd on the pty side you read/write */

	if (snagpty(pty,tty) == -1)  problemo("Owch! can't get a pty!");

	if((child=fork()) < 0) problemo("Owch! can't fork!");
	if(child) {
		/* parent */
#ifdef DEBUG
		fprintf(stderr,"PID %i Forked\r\n",child);
#endif
		return(child);
	} else {
		/* child */
		setsid();
		child=getpid();
		dup2(*tty,0);
		dup2(*tty,1);
		dup2(*tty, 2);

		execve("/bin/sh", argp, NULL);
		_exit(127);
	}
}

void fsock(char *cline)
{
	child=mypopen(&fdPTY,&fdTTY,cline);

	if(child==-1) {
		problemo("[UNABLE TO FORK LOCAL PROGRAM]");
	} else {
		pm.forked=1;
	}
}

int main(int argc,char *argv[])
{
	char commands[10240],arguement[10240];
	unsigned char buf[MAXIO];
	int length, ret, portno;
	fd_set rd;

	/* struct timeval ttime; */
	int i;
	int run=0;
	pm.passall=pm.forked=pm.recog=inflag=pm.passflag=outflag=pm.control=0;

	if( argc<3 ) {
		sprintf(commands,"\r\n%s\r\nUsage: %s host_address [port] [command junk]",TOOLVER,argv[0]);
		problemo(commands);
	}
	portno=PORT;
	commands[0]=0;
	for(i=2; i<argc;i++) {
		sprintf(arguement,"%s ",argv[i]);
		strcat(commands, arguement);
	}
	
	if(atoi(argv[2]) > 0) {
		commands[0]=0;
		for(i=3; i<argc;i++) {
			sprintf(arguement,"%s ",argv[i]);
			strcat(commands, arguement);
		}
		portno=((argc>=3 && atoi(argv[2])>0) ? atoi(argv[2]) : PORT);
	}
#ifdef DEBUG
	foo = fopen("blah", "a");
#endif

	signal(SIGPIPE,cclosed);
	signal(SIGCHLD,fireman);
	signal(SIGHUP,cclosed);

	setvbuf(stdin,NULL,_IONBF,0);
	setvbuf(stdout,NULL,_IONBF,0);

	sock=socket(AF_INET,SOCK_STREAM,0);
	server.sin_family=AF_INET;
	
	server.sin_port=htons(portno);
	server.sin_addr.s_addr=resolver(argv[1]);

	pm.progmode=((argc==4) || (argc==3 && atoi(argv[2])<=0));

	if(connect(sock,(struct sockaddr *)&server,sizeof(server))<0)
		problemo("Unable to connect to host.");
        setsockopt(sock,SOL_TCP,TCP_NODELAY,&flag, sizeof(int));
	start = time(NULL);

	for(;;) {
    
		errno=0;
		FD_ZERO(&rd);
		FD_SET(sock,&rd);
		if (pm.forked) FD_SET(fdPTY,&rd);

		errno = 0;
		ret = select(32,&rd,0,0,0);
		if(errno==EINTR)
			continue;
		/* PTY -> sock */
		if(pm.forked) {
			if (FD_ISSET(fdPTY,&rd)) {
				length=read(fdPTY,buf,MAXIO);
#ifdef DEBUG
				fprintf(stdout,"PTY -> sock Debug: ");
  
				do { ret = write(fileno(stdout),buf,length) } while( ret == -1 && errno==EINTR);
				fprintf(stdout,"\r\n");
#endif
				outpad(sock,buf,length,(unsigned int)255);
			}
		}
		/* sock -> PTY */
		if (pm.forked) {
			if(FD_ISSET(sock,&rd)) {
				if((length=read(sock,buf,MAXIO )) <= 0) cclosed();
				instrip(fdPTY,buf,length,(unsigned int)255);
#ifdef DEBUG
				fprintf(stdout,"sock -> PTY Debug: ");
				instrip(fileno(stdout),buf,length,(unsigned int)255);
				fprintf(stdout,"\r\n");
#endif
			}
		}

		if (run==0){
			run++;
			fsock(commands);
			sleep(5); /* build up que */
		} 
	}
}
